"""
    Geostreams Datapoints API
"""

from builtins import str
from builtins import object
import logging

from dateutil.parser import parse

from pygeotemporal.client import GeostreamsClient


class DatapointsApi(object):
    """
        API to manage the REST CRUD endpoints for datapoints.
    """
    def __init__(self, client=None, host=None, username=None, password=None):
        """Set client if provided otherwise create new one"""
        if client:
            self.api_client = client
        else:
            self.client = GeostreamsClient(host=host, username=username, password=password)

    def datapoint_post(self, datapoint):
        """
        Add a datapoint.

        :return: If successful or not.
        :rtype: `requests.Response`
        """
        logging.debug("Adding datapoint")
        try:
            return self.client.post("/datapoints", datapoint)
        except Exception as e:
            logging.error("Error adding datapoint %s: %s", datapoint, e.message)

    def datapoints_count_by_sensor_get(self, sensor_id):
        """
        Get the list of all available sensors.

        :return: Full list of sensors.
        :rtype: `requests.Response`
        """
        logging.debug("Counting datapoints by sensor")
        try:
            return self.client.get("/datapoints?sensor_id=%s&onlyCount=true" % sensor_id)
        except Exception as e:
            logging.error("Error counting datapoints by sensor %s: %s", sensor_id, e.message)

    def datapoint_latest_get(self, sensor_id, stream_id, since=None):
        """
        Get latest datapoint for a given stream by retrieving datapoint since a recent date and then grabbing
        the latest one.

        TODO: this should be an API endpoint
        """
        latest_datapoint = None
        # logging.info("Getting datapoints for stream %s since %s " % stream_id, since)
        if since is None:
            url = "/datapoints?stream_id=%s" % stream_id
        else:
            url = "/datapoints?stream_id=%s&since=%s" % (stream_id, since)
        try:
            datapoints = self.client.get_json(url)
            if isinstance(datapoints, list) and len(datapoints) > 0:
                latest_datapoint = datapoints[-1]
                start_date = parse(latest_datapoint['start_time']).strftime('%Y-%m-%d-%H-%M')
                logging.info("Fetched datapoints for " + str(sensor_id) + " starting on " + start_date)
            else:
                logging.debug("No datapoints exist for " + str(sensor_id))
        except Exception as e:
            logging.error("Error getting datapoints for stream %s since %s: %s", stream_id, since, e.message)

        return latest_datapoint

    def datapoint_create_json(self, start_time, end_time, longitude, latitude, sensor_id, stream_id, sensor_name,
                              properties=None, owner=None, source=None, procedures=None, elevation=0):
        datapoint = {
            'start_time': start_time,
            'end_time': end_time,
            'type': 'Feature',
            'geometry': {
                'type': "Point",
                'coordinates': [
                    longitude,
                    latitude,
                    elevation
                ]
            },
            'stream_id': str(stream_id),
            'sensor_id': str(sensor_id),
            'sensor_name': str(sensor_name)
        }

        if properties is not None:
            datapoint['properties'] = properties
            datapoint['properties']['site'] = sensor_name
            if owner is not None:
                datapoint['properties']["owner"] = owner
            if source is not None:
                datapoint['properties']['source'] = source
            if procedures is not None:
                datapoint['properties']['procedures'] = procedures

        return datapoint

    def datapoint_create_bulk(self, datapoints):
        logging.debug("Adding Datapoints in Bulk")
        try:
            return self.client.post("/datapoints/bulk", datapoints)
        except Exception as e:
            logging.error("Error adding bulk datapoint: %s", e.message)





# coding: utf-8

"""
    Geostreams Cache API
"""

from builtins import str
from builtins import object
import logging
from datetime import datetime

from dateutil import parser
from dateutil.relativedelta import relativedelta

from pygeotemporal.client import GeostreamsClient


class CacheApi(object):
    """
        This class provides functionality related to the Cache. Currently it is able to invalidate and prime the cache.
    """
    def __init__(self, client=None, host=None, username=None, password=None):
        """Set client if provided otherwise create new one"""
        if client:
            self.api_client = client
        else:
            self.client = GeostreamsClient(host=host, username=username, password=password)

    def invalidate_cache(self, sensor_id):
        """Invalidate sensor cache"""
        logging.debug("Invalidating cache for sensor %s" % sensor_id)
        try:
            return self.client.get("/cache/invalidate?sensor_id=%s" % sensor_id)
        except Exception as e:
            logging.error("Error invalidating cache for sensor %s" % sensor_id, e.message)

    def bin(self, sensor_id):
        """Bin sensor's datapoints"""
        logging.debug("Binning %s" % sensor_id)
        try:
            return self.client.post("/cache?sensor_id=%s" % sensor_id, content=None)
        except Exception as e:
            logging.error("Error binning sensor %s" % sensor_id, e.message)

    def prime_cache(self, sensor_id):
        """Prime the cache of a specific sensor."""
        logging.info(" Caching datapoints for sensor with id: " + str(sensor_id))

        try:
            sensor = self.client.get("/sensors/%s" % str(sensor_id)).json()
        except Exception as e:
            logging.error("Error getting sensor %s" % sensor_id, e.message)

        # Calculate time in days between start and end date of sensor info
        delta_days = (datetime.strptime(sensor['max_end_time'], "%Y-%m-%dT%H:%M:%SZ") -
                      datetime.strptime(sensor['min_start_time'], "%Y-%m-%dT%H:%M:%SZ")).days

        # Calculate time in years between start and end date of sensor info
        delta_years = relativedelta(parser.parse(sensor['max_end_time']), parser.parse(sensor['min_start_time'])).years

        if delta_years > 100:
            time_bin = 'decade'
        elif delta_years > 50:
            time_bin = 'lustrum'
        elif delta_years > 25:
            time_bin = 'year'
        elif delta_years > 10:
            time_bin = 'season'
        elif delta_years > 1:
            time_bin = 'month'
        else:
            if delta_days > 14:
                time_bin = 'day'
            elif delta_days > 3:
                time_bin = 'hour'
            else:
                time_bin = 'minute'

        try:
            return self.client.get("/datapoints/bin/" + str(time_bin) + "/1?sensor_id=" + str(sensor_id))
        except Exception as e:
            logging.error("Error priming cache for sensor %s" % sensor_id, e.message)



from __future__ import print_function
from builtins import str
import requests
import json
import argparse


def main():
    """
    This function retrieves all sensors from the host, and generates or updates all bins for them.
    The user needs to have master permissions in the geostreams service to be able to generate the bins.\

    Arguments:
        host - the url for the geostreams service. Example: http://seagrant-dev.ncsa.illinois.edu:9002/geostreams
        username - user with master permission in geostream service
        password - password for user in geostream service
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("--host", help="host url",
                        default="http://localhost:9000/")
    # Sample URL to go around nginx
    # - nginx times out the request because it takes longer than 5 seconds
    # - "http://seagrant-dev.ncsa.illinois.edu:9002/geostreams")
    parser.add_argument('-u', '--user', help="Geostreams username")
    parser.add_argument('-p', '--password', help="Geostreams password")

    args = parser.parse_args()
    host = args.host
    username = args.user
    password = args.password

    user = {'identifier': username, 'password': password}
    r = requests.post(host + '/api/authenticate',
        data=json.dumps(user), headers={'Content-Type':'application/json'})

    if r.status_code == 200:
        r.headers["Content-Encoding"] = "application/json"

        sensors = requests.get(host + "/api/sensors")
        sensors_json = json.loads(sensors.text)
        parsed_sensors = sensors_json["sensors"]

        for sensor in parsed_sensors:
            sensor_id = sensor["id"]
            endpoint = host + '/api/cache?sensor_id=' + str(sensor_id)
            s = requests.post(endpoint, headers=r.headers)
            print("Cache created for sensor with id: " + str(sensor_id))


if __name__ == "__main__":
    main()



# coding: utf-8

"""
    Parser scheduler.
"""

from builtins import range
import schedule
import subprocess
import smtplib
import time
from email.mime.text import MIMEText


def jobexecute(execution, title, toaddress, sendemail):
    try:
        # for test
        # print "I'm working..."
        subprocess.check_output(execution, shell=True, stderr=subprocess.STDOUT)
        return 1
    except subprocess.CalledProcessError as e: 
        if sendemail:
            now = time.strftime("%c")
            fromaddress = "devnull@ncsa.illinois.edu"
            message = MIMEText("Failed at %s \n-----------------------------------\n %s \n" % (now, e.output))
            message['Subject'] = title
            message['From'] = fromaddress
            message['To'] = toaddress

            server = smtplib.SMTP('smtp.ncsa.illinois.edu', 25)
            server.sendmail(fromaddress, toaddress.split(", "), message.as_string())
            server.quit()

        return 0


# re-execute the job for executionnumber times, and send email the last time. sleep for 10 mins between each execution.
def job(execution, title, toaddress, executionnumber):
    for i in range(0, executionnumber - 1):
        execute_status = jobexecute(execution, title, toaddress, False)
        if execute_status == 1:
            return 
        else:
            time.sleep(10*60)
    jobexecute(execution, title, toaddress, True)
    return 


def scheduler(execution, title, toaddress, executionnumber): 
    # https://github.com/dbader/schedule
    # for test
    # scheduler = schedule.every(1).minutes
    scheduler = schedule.every().monday
    kwargs = {"execution": execution, "title": title, "toaddress": toaddress, "executionnumber": executionnumber}
    scheduler.do(job, **kwargs)

    while True:
        schedule.run_pending()
        time.sleep(1)



"""
    Geostreams Client
    ~~~
    This module contains a basic client to interact with the Geostreams API.
"""


from builtins import object
import json
import logging
import time
import requests


class GeostreamsClient(object):
    """
    Client to Geostreams API to store connection information.

    The `path` parameter used by many of the methods in this class call a specific path relative to the host + "api".
    For example passing in "/version" for host "https://seagrant-dev.ncsa.illinois.edu/" will call
    "https://seagrant-dev.ncsa.illinois.edu/api/version". Make sure to include the slash at the beginning of
    the fragment.
    """
    logger = logging.getLogger(__name__)
    api_fragment = "/api"
    max_retries = 10
    call_timeout = 5

    def __init__(self, *args, **kwargs):
        """
        Create an instance of `GeostreamsClient`.

        :param string host: The root host url of the specific geostreaming API we are connecting to.
        :param string key: The API key used to write to the API. Set this or `username`/`password` below.
        :param string username: HTTP Basic Authentication username.
        :param string password: HTTP Basic Authentication password.
         """
        self.host = kwargs.get('host')
        self.username = kwargs.get('username')
        self.password = kwargs.get('password')
        self.user = {'identifier': self.username, 'password': self.password}

        login_url = self.host + "/api/authenticate"
        r = requests.post(login_url, json=self.user, headers={'Content-Type': 'application/json'})

        self.headers = {"x-auth-token": r.headers["x-auth-token"], "Content-Encoding": "application/json"}

    def version(self):
        """Return Geostreams version info."""
        url = self.host + self.api_fragment + "/version"
        self.logger.debug("GET %s", url)
        r = requests.get(url)
        if r.status_code == requests.codes.ok:
            try:
                json = r.json()
                self.logger.debug("Version: %s", json)
                return json
            except ValueError:
                self.logger.error("GET %s. Could not parse JSON. Status %s.", url, r.status_code)
                r.raise_for_status()
        else:
            r.raise_for_status()

    def get_json(self, path):
        """
        Call HTTP GET against `path`. This version returns a JSON object.

        :param path: Endpoint path relative to Geostreams api.
        :return: JSON from body of response
        :rtype: JSON
        :raises: `requests.HTTPError`
        """
        url = self.host + self.api_fragment + path
        r = requests.get(url, headers=self.headers)
        r.raise_for_status()
        return r.json()

    def get(self, path):
        """
        Call HTTP GET against `path`. This version returns a `requests.Response` object.

        :param path: Endpoint path relative to Geostreams api.
        :return: Full response object so that we can check status on it and then retrieve the JSON body.
        :rtype: `requests.Response`
        :raises: `requests.HTTPError`
        """
        url = self.host + self.api_fragment + path
        try:
            return requests.get(url, headers=self.headers)
        except Exception as e:
            logging.exception("Error calling GET url %s: %s" % url, e.message)

    def get_retry(self, path):
        """
        Call HTTP GET against `path`. This version returns a `requests.Response` object. Useful in case of temporary
        network issues.

        :param path: Endpoint path relative to Geostreams api.
        :return: Full response object so that we can check status on it and then retrieve the JSON body.
        :rtype: `requests.Response`
        :raises: `requests.HTTPError`
        """
        url = self.host + self.api_fragment + path
        count = 0
        while True:
            try:
                if count > self.max_retries:
                    return None
                r = requests.get(url, headers=self.headers)
                count += 1
                if 200 <= r.status_code < 300:
                    return r.json()
                else:
                    logging.warning("Error calling GET url %s" % url)
                    logging.warning("Waiting %i seconds and will try again" % self.call_timeout)
                    time.sleep(self.call_timeout)
            except Exception as e:
                logging.exception("Error calling GET url %s: %s" % url, e.message)

    def post(self, path, content):
        """
        Call HTTP POST against `path` with `content` in body.

        :param path: Endpoint path relative to Geostreams api.
        :param content: Content to send as the body of the request.
        :return: Full response object so that we can check status on it and then retrieve the JSON body.
        :rtype: `requests.Response`
        :raises: `requests.HTTPError`
        """
        url = self.host + self.api_fragment + path

        try:
            k = requests.post(url, json=content, headers=self.headers)
            return k
        except Exception as e:
            self.logger.error("POST %s: %s", url, e.message)

    def put(self, path, content):
        """
        Call HTTP POST against `path` with `content` in body.

        :param path: Endpoint path relative to Geostreams api.
        :param content: Content to send as the body of the request.
        :return: Full response object so that we can check status on it and then retrieve the JSON body.
        :rtype: `requests.Response`
        :raises: `requests.HTTPError`
        """
        url = self.host + self.api_fragment + path

        try:
            k = requests.put(url, json=content, headers=self.headers)
            return k
        except Exception as e:
            self.logger.error("PUT %s: %s", url, e.message)

    def post_file(self, path, filename):
        """
        Call HTTP POST against `path` with `content` in body. Header with content-type is not required.

        :param path: Endpoint path relative to Geostreams api.
        :param content: Content to send as the body of the request.
        :return: Full response object so that we can check status on it and then retrieve the JSON body.
        :rtype: `requests.Response`
        :raises: `requests.HTTPError`
        """

        url = self.host + self.api_fragment + path

        try:
            return requests.post(url, files={"File": open(filename, 'rb')},
                                 headers=self.headers)
        except Exception as e:
            self.logger.error("POST %s: %s", (url, e.message))

    def post_retry(self, path, content):
        """
        Call HTTP POST against `path` with `content` in body. Retry up to a certain number of times if necessary. Useful
        in case of temporary network issues.

        :param path: Endpoint path relative to Geostreams api.
        :param content: Content to send as the body of the request.
        :return: Full response object so that we can check status on it and then retrieve the JSON body.
        :rtype: `requests.Response`
        :raises: `requests.HTTPError`
        """
        url = self.host + self.api_fragment + path
        count = 0
        while True:
            try:
                if count > self.max_retries:
                    return None
                r = requests.post(url, data=json.dumps(content), headers=self.headers)
                count += 1
                if 200 <= r.status_code < 300:
                    return r.json()
                else:
                    logging.warning("Error calling POST url %s" % url)
                    logging.warning("Waiting %i seconds and will try again" % self.call_timeout)
                    time.sleep(self.call_timeout)
            except Exception as e:
                logging.exception("Error calling POST url %s: %s" % url, e.message)

    def delete(self, path):
        """
        Call HTTP DELETE against `path`.

        :param path: Endpoint path relative to Geostreams api.
        :return: Full response object so that we can check status on it and then retrieve the JSON body.
        :rtype: `requests.Response`
        :raises: `requests.HTTPError`
        """
        url = self.host + self.api_fragment + path

        try:
            return requests.delete(url, headers=self.headers)
        except Exception as e:
            self.logger.error("DELETE %s: %s", url, e.message)

    def delete_retry(self, path):
        """
        Call HTTP DELETE against `path`. Retry up to a certain number of times if necessary. Useful in case of temporary
        network issues.

        :param path: Endpoint path relative to Geostream api.
        :return: Full response object so that we can check status on it and then retrieve the JSON body.
        :rtype: `requests.Response`
        :raises: `requests.HTTPError`
        """
        url = self.host + self.api_fragment + path
        count = 0
        while True:
            try:
                if count > self.max_retries:
                    return None
                r = requests.delete(url, headers=self.headers)
                count += 1
                if 200 <= r.status_code < 300:
                    return r.json()
                else:
                    logging.warning("Error calling DELETE url %s" % url)
                    logging.warning("Waiting %i seconds and will try again" % self.call_timeout)
                    time.sleep(self.call_timeout)
            except Exception as e:
                logging.exception("Error calling DELETE url %s: %s" % url, e.message)



from datetime import datetime, timedelta

from pytz import timezone, utc


# CONVERT TO UTC FORMAT            FROM EXAMPLE FORMAT 2015-11-20 03:46:20
def time2utc(date_in, time_zone="America/Chicago"):
    date_out = timezone(time_zone).localize(datetime.strptime(date_in, "%Y-%m-%d %H:%M:%S")).astimezone(utc).strftime(
        "%Y-%m-%dT%H:%M:%SZ")

    return date_out


# convert year and julian day to date
def julian_day_to_month_day(year,julian_day):
    date = datetime(int(year), 1, 1) + timedelta(int(julian_day))

    return date


# convert calendar date month/day/2-digit-year to YYYY-MM-DDT00:00:00Z (assumes there is no time so all set to zero
def calendar_date2utc(date_in,time_zone="America/Chicago"):

    date_out = datetime.strptime(date_in + " 00:00:00", "%m/%d/%y %H:%M:%S").strftime("%Y-%m-%dT%H:%M:%SZ")

    return date_out



"""
    Geostreams Streams API
"""

from builtins import range
from builtins import object
import logging
import json
from pygeotemporal.client import GeostreamsClient


class StreamsApi(object):
    """
        API to manage the REST CRUD endpoints for Streams.
    """
    def __init__(self, client=None, host=None, username=None, password=None):
        """Set client if provided otherwise create new one"""
        if client:
            self.api_client = client
        else:
            self.client = GeostreamsClient(host=host, username=username, password=password)

    def streams_get(self):
        """
        Get the list of all available streams.

        :return: Full list of streams.
        :rtype: `requests.Response`
        """
        logging.debug("Getting all streams")
        try:
            return self.client.get("/streams")
        except Exception as e:
            logging.error("Error retrieving stream list: %s", e.message)

    def streams_get_by_id(self, id):
        logging.debug("Getting stream with id %s" % id)
        try:
            return self.client.get("/streams/%s" % id)
        except Exception as e:
            logging.error("Error retrieving stream with id %s: %s" % id, e.message )

    def stream_get_by_name_json(self, stream_name):
        """
        Get a specific stream by id.

        :return: stream object as JSON.
        :rtype: `requests.Response`
        """
        logging.debug("Getting stream %s" % stream_name)
        stream = self.client.get("/streams?stream_name=" + stream_name).json()
        if 'status' in stream and stream['status'] == "No data found" or not stream["streams"]:
            return None
        else:
            return stream

    def stream_post(self, stream):
        """
        Create stream.

        :return: stream json.
        :rtype: `requests.Response`
        """
        logging.debug("Adding stream")

        try:
            return self.client.post("/streams", stream)
        except Exception as e:
            logging.error("Error retrieving streams: %s", e.message)

    def stream_post_json(self, stream):
        """
        Create stream.

        :return: stream json.
        :rtype: `requests.Response`
        """
        logging.debug("Adding or getting stream")

        stream_from_geostreams = self.stream_get_by_name_json(stream['name'])

        if stream_from_geostreams is None:
            logging.info("Creating stream with name: " + stream['name'])
            result = self.client.post("/streams", stream)
            if result.status_code != 200:
                logging.info("Error posting stream")
            stream_from_geostreams = self.stream_get_by_name_json(stream['name'])

        logging.info("Found stream %s", stream['name'])
        return stream_from_geostreams["streams"][0]

    def stream_delete(self, stream_id):
        """
        Delete a specific stream by id.

        :return: If successfull or not.
        :rtype: `requests.Response`
        """
        logging.debug("Deleting stream %s" % stream_id)
        try:
            return self.client.delete("/streams/%s" % stream_id)
        except Exception as e:
            logging.error("Error deleting stream %s: %s", stream_id, e.message)

    def stream_create_json_from_sensor(self, sensor):
        """
        Create stream from sensor. Note: It does not post the stream to the API

        :param: sensor
        :return: stream Json object
        """

        stream = {
            "sensor_id": sensor['id'],
            "name": sensor['name'],
            "type": sensor['geoType'],
            "geometry": sensor['geometry'],
            "properties": sensor['properties']
        }

        return stream

    def stream_delete_range(self, start, end, keyword):
        """
        Deletes streams in a range of indexes [start, end] where the name includes keyword.

        """

        for i in range(start, end + 1):
            stream = self.streams_get_by_id(i)
            json_stream = json.loads(stream.content)
            if 'name' in json_stream:

                if keyword.lower() in json_stream['name'].encode("ascii").lower():
                    self.stream_delete(i)
                    logging.info("Stream Deleted, %s", i)
                else:
                    logging.info("Sensor not deleted %s, no keyword match, stream name: %s", i, json_stream["name"])

            else:
                logging.info("No name keyword in stream, stream doesn't exist. Stream Id: %s", i)
        return



from LatLon import LatLon
from LatLon import Latitude
from LatLon import Longitude

"""Convert a gecode 2d list from [[deg,min,sec],[deg,min,sec]] to [decimal,decimal] """


def dms2dec(dms):
    dec = list(LatLon(Latitude(degree=dms[0][0], minute=dms[0][1], second=dms[0][2]),
                      Longitude(degree=dms[1][0], minute=dms[1][1], second=dms[1][2])).to_string())

    return [float(dec[0]), float(dec[1])]



# coding: utf-8

"""
    Load data from CSV files.
"""
from __future__ import absolute_import
from builtins import range
from builtins import object
import logging

from pygeotemporal.time_transformers import time2utc


def load(file):
    pass


class ParametersMap(object):
    def __init__(self, yaml_file):
        pass

    def lookup(self, csv_name):
        return "new_parameter_id"


def parameters_map(yaml_file):
    pass


def get_needed_values(needed_names, column_headers_in, parameters_map, data_in):
    """Drop all columns from data_in that are not in needed_names

    Keyword arguments:
        needed_names -- array of columns to include
        column_headers_in -- array of all columns from original file
        parameters_map -- `ParametersMap`
        data_in -- 2D array for data values from original file
    """
    column_headers_out = []
    needed_indexes = []
    for i in range(len(column_headers_in)):
        if column_headers_in[i] in needed_names:
            column_headers_out.append(parameters_map.lookup(column_headers_in[i]))
            needed_indexes.append(i)

    logging.info(" [pyG.map_names] Removing unneeded data values.")
    data_hold = []
    while len(data_in) > 0:
        data_row = []
        for i in needed_indexes:
            data_row.append(data_in[0][i])
        data_hold.append(data_row)
        data_in.pop(0)

    data_out = []
    while len(data_hold) > 0:
        data_dict = {}
        # FIXME why return this data dictionary?
        data_dict['sensor_name'] = data_hold[0][column_headers_out.index("sensor_name")]
        data_dict['lat'] = float(data_hold[0][column_headers_out.index("lat")])
        data_dict['long'] = float(data_hold[0][column_headers_out.index("long")])
        data_dict['start_time'] = time2utc(data_hold[0][column_headers_out.index("start_date")])
        data_dict['time_zone'] = data_hold[0][column_headers_out.index("time_zone")]

        parameter_name = parameters_map.lookup(data_hold[0][column_headers_out.index("Characteristic Name")],
                                               data_hold[0][column_headers_out.index("Sample Fraction")])

        try:
            pass
            data_dict[parameter_name] = float(data_hold[0][column_headers_out.index("Result Value as Number")])
        except:
            logging.warning("[pyG.map_names] couldn't get value for datapoint - skipping datapoint")
            data_hold.pop(0)
            continue
        data_out.append(data_dict)
        data_hold.pop(0)

    return data_out



from distutils.core import setup

setup(name='pygeotemporal',
      version='0.1.0',
      url='https://opensource.ncsa.illinois.edu/bitbucket/projects/GEOD/repos/pygeotemporal/browse',
      packages=['pygeotemporal'],
      )



"""
    Geostreams Sensors API.
"""
from __future__ import division

from builtins import object
import logging
import time
from datetime import datetime

from pygeotemporal.client import GeostreamsClient


class SensorsApi(object):
    """
        API to manage the REST CRUD endpoints for sensors.
    """
    def __init__(self, client=None, host=None, username=None, password=None):
        """Set client if provided otherwise create new one"""
        if client:
            self.api_client = client
        else:
            self.client = GeostreamsClient(host=host, username=username, password=password)

    def sensors_get(self):
        """
        Get the list of all available sensors.

        :return: Full list of sensors.
        :rtype: `requests.Response`
        """
        logging.debug("Getting all sensors")
        try:
            return self.client.get("/sensors")
        except Exception as e:
            logging.error("Error retrieving sensor list: %s", e.message)

    def sensors_refresh_cache(self):
        """
        Get all available sensors one by one, return nothing

        :rtype: `requests.Response`
        """
        logging.debug("Getting all sensors by each id")
        try:
            sensor_list = self.client.get("/sensors")

            for sensor in sensor_list.json():
                logging.debug("refresh sensor %s", sensor['id'])
                self.client.get("/sensors/%s" % sensor['id'])
                binning_filter = self.get_aggregation_bin(sensor['min_start_time'], sensor['max_end_time'])
                if sensor['properties']['type']['id'] == 'epa': 
                    binning_filter = 'semi'

                self.client.get("/datapoints/bin/%s/1?sensor_id=%s" % (binning_filter, sensor['id']))
                time.sleep(100)     
            return True    
        except Exception as e:
            logging.error("Error caching sensor list: %s", e.message)

    def sensor_refresh_cache(self, sensor_id):
        """
        Get all available sensors one by one, return nothing

        :rtype: `requests.Response`
        """
        logging.debug("prime cache for sensor %s" % sensor_id)
        try:
            response = self.client.get("/sensors/%s" % sensor_id)
            sensor = response.json()
            if 'min_start_time' not in sensor: 
                return False
            logging.debug("refresh sensor %s", sensor_id)
            self.client.get("/sensors/%s" % sensor_id)
            binningfilter = self.get_aggregation_bin(sensor['min_start_time'], sensor['max_end_time'])
            if sensor['properties']['type']['id'] == 'epa': 
                binningfilter = 'semi'

            self.client.get("/datapoints/bin/%s/1?sensor_id=%s" % (binningfilter, sensor_id))
    
            return True    
        except Exception as e:
            logging.error("Error caching sensor: %s", e.message)

    def sensor_get(self, sensor_id):
        """
        Get a specific sensor by id.

        :return: Sensor object as JSON.
        :rtype: `requests.Response`
        """
        logging.debug("Getting sensor %s" % sensor_id)
        try:
            return self.client.get("/sensors/%s" % sensor_id)
        except Exception as e:
            logging.error("Error retrieving sensor %s: %s", sensor_id, e.message)

    def sensor_get_by_name(self, sensor_name):
        """
        Get a specific sensor by id.

        :return: Sensor object as JSON.
        :rtype: `requests.Response`
        """
        logging.debug("Getting sensor %s" % sensor_name)
        try:
            return self.client.get("/sensors?sensor_name=" + sensor_name)
        except Exception as e:
            logging.error("Error retrieving sensor %s: %s", sensor_name, e.message)
            return None

    def sensor_post(self, sensor):
        """
        Create sensor.

        :return: sensor json.
        :rtype: `requests.Response`
        """
        logging.debug("Adding sensor")
        try:
            return self.client.post("/sensors", sensor)
        except Exception as e:
            logging.error("Error adding sensor %s: %s", sensor['name'], e.message)

    def sensor_post_json(self, sensor):
        """
        Create sensor.

        :return: sensor json.
        :rtype: `requests.Response`
        """
        logging.debug("Adding or getting sensor")
        try:

            if 'id' in sensor:
                sensor_from_geostreams = self.sensor_get(sensor['id']).json()
            elif 'name' in sensor:
                sensor_from_geostreams = self.sensor_get_by_name(sensor['name']).json()

            if len(sensor_from_geostreams['sensors']) > 0:
                logging.info("Found sensor " + sensor['name'])
                return sensor_from_geostreams['sensors'][0]
            else:
                logging.info("Creating sensor with name: " + sensor['name'])
                sensor_id = (self.client.post("/sensors", sensor)).json()
                sensor_from_geostreams = self.sensor_get(sensor_id['id']).json()
                return sensor_from_geostreams['sensor']
        except Exception as e:
            logging.error("Error adding sensor %s: %s", sensor['name'], e.message)

    def sensor_delete(self, sensor_id):
        """
        Delete a specific sensor by id.

        :return: If successful or not.
        :rtype: `requests.Response`
        """
        logging.debug("Deleting sensor %s" % sensor_id)
        try:
            return self.client.delete("/sensors/%s" % sensor_id)
        except Exception as e:
            logging.error("Error retrieving sensor %s: %s", sensor_id, e.message)

    def sensor_create_json(self, name, longitude, latitude, elevation, popup_content, region, huc=None, network=None,
                           organization_id=None, title=None):
        """Create sensor definition in JSON"""
        sensor = {
            "name": name,
            "type": "Feature",
            "geometry": {
                "type": "Point",
                "coordinates": [
                    longitude,
                    latitude,
                    elevation
                ]
            },
            "properties": {
                "name": name,
                "popupContent": popup_content,
                "region": region
            }
        }

        if network or id or title:
            sensor['properties']['type'] = {}

        if huc:
            sensor["properties"]["huc"] = huc
        if network:
            sensor['properties']['type']['network'] = network
        if organization_id:
            sensor['properties']['type']['id'] = organization_id
        if title:
            sensor['properties']['type']['title'] = title
        return sensor

    def sensor_statistics_post(self, sensor_id):
        """
        Update sensor statistics.

        :return: Full list of sensors.
        :rtype: `requests.Response`
        """
        logging.debug("Updating sensor statistics")
        sensor = (self.sensor_get(sensor_id)).json()["sensor"]
        try:
            return self.client.put("/sensors/%s/update" % sensor_id, content=sensor)
        except Exception as e:
            logging.error("Error updating sensor statistics for sensor %s: %s", sensor_id, e.message)

    def get_aggregation_bin(self, start_time, end_time):
        start_time = datetime.strptime(start_time, '%Y-%m-%dT%H:%M:%SZ')
        end_time = datetime.strptime(end_time, '%Y-%m-%dT%H:%M:%SZ')
        year = end_time.year - start_time.year

        if year > 100: 
            time_bin = 'decade'
        elif year > 50: 
            time_bin = 'lustrum'
        elif year > 25:
            time_bin = 'year'
        elif year > 10:
            time_bin = 'season'
        elif year > 1:
            time_bin = 'month'
        else:
            diff_time = end_time - start_time
            b_time = divmod(diff_time.days * 86400 + diff_time.seconds, 60)[0]  # minutes
            b_time = b_time / 1440
            if b_time > 14:
                time_bin = 'day'
            elif b_time > 3:
                time_bin = 'hour'
            else:
                time_bin = 'minute'
        return time_bin

    def update_sensor_metadata(self, sensor):
        """
        Update sensor metadata
        :param sensor: Sensor json
        :return: Http response
        """
        logging.debug("Updating sensor's metadata")
        sensor_id = sensor["id"]
        try:
            return self.client.put("/sensors/%s" % sensor_id, content=sensor)
        except Exception as e:
            logging.error("Error updating sensor metadata for sensor %s: %s", sensor_id, e.message)
